import pathlib

from .static import StaticRule, build_target

from irori.context import build
from irori.context import runtime
from irori.executors import Executor


class LinkRule(StaticRule):
    def __init__(self, command=None, debug=None, release=None, rules=None):
        super(LinkRule, self).__init__(
            command=command, debug=debug, release=release, output_name_converter=build_target)
        if rules is None:
            rules = []
        elif not isinstance(rules, list):
            rules = [rules]
        self.__rules = rules

    def initialize(self, context: build.BuildContext):
        super(LinkRule, self).initialize(context)
        for rule in self.__rules:
            rule.initialize(context)

    def execute(self, context: runtime.RuntimeContext, executor: Executor):
        for rule in self.__rules:
            outputs = rule.execute(context, executor)
            executor.wait_all()
            context = runtime.RuntimeContext([pathlib.Path(output) for output in outputs])
        outputs = super(LinkRule, self).execute(context, executor)
        executor.wait_all()

        return outputs



import pathlib


def get_modified_time(file):
    file = pathlib.Path(file)
    if file.exists():
        return file.stat().st_mtime
    return 0


class Dependencies:
    def __init__(self, *depend_files_name, files=None):
        if files is None:
            files = []

        files.extend(list(depend_files_name))

        self.__dependencies = Dependencies.__resolve_depend(files)

    @staticmethod
    def __resolve_depend(files, deps=None):
        if deps is None:
            deps = {}
        for file in files:
            with open(file) as fp:
                for line in fp.readlines():
                    line = line.split(':')
                    out = line[0].strip()
                    inp = [i.strip() for i in line[1].strip().split(' ')]
                    if out not in deps:
                        deps[out] = []
                    deps[out].extend(inp)
        return deps

    def latest_time(self, output):
        times = [get_modified_time(d) for d in self.depend_files(output)]
        if len(times) == 0:
            return 0
        return max(times)

    def depend_files(self, output):
        if output not in self.__dependencies:
            return []

        deps = self.__dependencies[output]
        for dep in deps:
            deps.extend(self.depend_files(dep))

        return deps



import pathlib
import enum
import argparse

from . import base
from irori.depend import Dependencies


class Mode(enum.Enum):
    debug = 'debug'
    release = 'release'


class BuildContext(base.Context):
    def __init__(self,
                 obj: pathlib.Path,
                 output: pathlib.Path, mode: Mode, dependencies: Dependencies, args: argparse.Namespace):
        self.__object_dir = obj
        self.__output_file = output
        self.__mode = mode
        self.__dependencies = dependencies
        self.__args = args

    @property
    def object_directory(self) -> pathlib.Path:
        return self.__object_dir

    @property
    def output_file(self) -> pathlib.Path:
        return self.__output_file

    @property
    def mode(self) -> Mode:
        return self.__mode

    @property
    def is_debug(self) -> bool:
        return self.mode == Mode.debug

    @property
    def is_release(self) -> bool:
        return self.mode == Mode.release

    @property
    def dependencies(self) -> Dependencies:
        return self.__dependencies

    @property
    def arguments(self) -> argparse.Namespace:
        return self.__args



import typing
import pathlib

from .base import Context


class RuntimeContext(Context):
    def __init__(self, sources: typing.List[pathlib.Path]):
        self.__sources = sources
        self.__original_sources = sources
        self.__current = None
        self.__next = self.__sources[0]

    def consume(self, count=1) -> typing.List[pathlib.Path]:
        self.__current = self.__sources[:count]
        self.__sources = self.__sources[count:]
        if self.is_remaining:
            self.__next = self.__sources[0]
        else:
            self.__next = None
        return self.__current

    def consume_all(self) -> typing.List[pathlib.Path]:
        sources = self.__sources
        self.__sources = []
        return sources

    @property
    def remaining_count(self) -> int:
        return len(self.__sources)

    @property
    def is_remaining(self):
        return self.remaining_count > 0

    @property
    def remaining(self):
        return self.__sources

    @property
    def sources(self):
        return self.__original_sources

    @property
    def current(self):
        return self.__current

    @property
    def next(self):
        return self.__next



import sys

from irori.executors import Executor
from .base import Rule
from irori.context import build
from irori.context import runtime


class SuffixRule(Rule):
    def __init__(self, rules=None, **kw_rules):
        super(SuffixRule, self).__init__()
        if rules is None:
            rules = {}
        rules.update(kw_rules)

        self.__rules = {suffix.strip('/\\.'): rule for suffix, rule in rules.items()}

    def initialize(self, context: build.BuildContext):
        super(SuffixRule, self).initialize(context)
        for rule in self.__rules.values():
            rule.initialize(context)

    def execute(self, context: runtime.RuntimeContext, executor: Executor):
        outputs = []
        while context.is_remaining:
            suffix = context.next.suffix.strip('/\\.')
            if suffix in self.__rules:
                outputs.extend(self.__rules[suffix].execute(context, executor))
            else:
                print('unknown suffix: {}, file name: {}'.format(suffix, context.consume()[0]), file=sys.stderr)
        executor.wait_all()

        return outputs



import abc

from irori.context import build
from irori.context import runtime
from irori.executors import Executor


class Rule(abc.ABC):
    def __init__(self):
        self.__build_context = None

    def initialize(self, context: build.BuildContext):
        self.__build_context = context

    @property
    def context(self) -> build.BuildContext:
        return self.__build_context

    @abc.abstractmethod
    def execute(self, context: runtime.RuntimeContext, executor: Executor):
        pass



import argparse
import pathlib
import multiprocessing

from irori import context, executors
from irori.depend import Dependencies


def start(target, recipes, finder, dependencies=None, arguments=None):
    parser = argparse.ArgumentParser(description='Build tool')
    parser.add_argument('--debug', '-d', help='run as debug mode', action='store_true')
    parser.add_argument('--release', '-r', help='run as release mode (default)', action='store_true')
    parser.add_argument(
        '--object', '--obj', '--object-directory', help='temporary object output directory (default obj)',
        default='obj')
    parser.add_argument('--no-concurrency', help='disable concurrency', action='store_true')
    parser.add_argument(
        '--concurrency',
        help='concurrency level > 0 (default {})'.format(multiprocessing.cpu_count()),
        type=int, default=multiprocessing.cpu_count())

    handler = None

    if dependencies is None:
        dependencies = Dependencies()

    if arguments is not None:
        arguments(parser)
    recipe_entries = recipes()

    sub = parser.add_subparsers()
    subparsers = {recipe: sub.add_parser(recipe) for recipe in recipe_entries.keys()}
    [parser.set_defaults(handler=recipe_entries[name]) for name, parser in subparsers.items()]

    args = parser.parse_args()
    if hasattr(args, 'handler'):
        handler = args.handler

    if handler is None:
        parser.print_help()
        exit(1)

    build_context = context.BuildContext(
        pathlib.Path(args.object),
        target, context.Mode.debug if args.debug else context.Mode.release, dependencies, args)
    rule = handler(build_context)

    dependencies = finder(build_context)

    core = args.concurrency
    if args.no_concurrency:
        core = 1

    executor = executors.Executor(core)

    rule.initialize(build_context)

    runtime_context = context.RuntimeContext([pathlib.Path(dep) for dep in dependencies])

    rule.execute(runtime_context, executor)
    executor.wait_all()



import pathlib
import typing

from . import base
from irori.context import build
from irori.context import runtime
from irori.executors import Executor
from irori import depend


def _convert_command_2nd(command: str):
    return command.replace('$@', '{output}')


def _convert_command_1st(command: str, default):
    if command is None:
        return default
    return command.replace('$<', '{dependency}').replace('$^', '{dependencies}')


def _latest_deps_time(output, dependencies: typing.List[pathlib.Path], deps: depend.Dependencies):
    lt = deps.latest_time(output)
    return max(lt, max([depend.get_modified_time(str(f)) for f in dependencies]))


def single_output_name_converter(context: build.BuildContext, dependencies: typing.List[pathlib.Path]):
    context.object_directory.mkdir(parents=True, exist_ok=True)
    return '{}/{}.o'.format(context.object_directory, dependencies[0].name)


def build_target(context: build.BuildContext, _: typing.List[pathlib.Path]):
    return context.output_file


class StaticRule(base.Rule):
    def __init__(
            self, command=None, debug=None, release=None,
            output_name_converter=single_output_name_converter, force=False):
        super(StaticRule, self).__init__()
        self.__command = _convert_command_1st(command, '')
        self.__debug = _convert_command_1st(debug, '')
        self.__release = _convert_command_1st(release, '')
        self.__output_name_converter = output_name_converter
        self.__force = force

    def initialize(self, context: build.BuildContext):
        super(StaticRule, self).initialize(context)
        self.__command += ' '
        self.__command += self.__debug if context.is_debug else self.__release

    def execute(self, context: runtime.RuntimeContext, executor: Executor):
        command = self.__command
        if '{dependency}' in command:
            dependencies = context.consume()
            command = command.format(dependency=dependencies[0])
        elif '{dependencies}' in command:
            dependencies = context.consume_all()
            command = command.format(dependencies=' '.join([str(dep) for dep in dependencies]))
        else:
            dependencies = []
            command = command
        command = _convert_command_2nd(command)

        output = self.__output_name_converter(self.context, dependencies)

        if _latest_deps_time(output, dependencies, self.context.dependencies) > depend.get_modified_time(output):
            command = command.format(output=output)
            executor.start_process(command)
            print('execute: {}'.format(command))
        else:
            print('skipped: {}'.format(output))

        return [output]



import subprocess


class Executor:
    def __init__(self, concurrency_count=1):
        self.__processes = []
        self.__concurrency_count = concurrency_count
        self.__running = True

    def start_process(self, command: str):
        if not self.__running:
            return None
        if len(self.__processes) >= self.__concurrency_count:
            pass
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        self.__processes.append(process)
        return process

    @property
    def concurrency_count(self):
        return self.__concurrency_count

    @staticmethod
    def __process_failed(proc: subprocess.Popen):
        stdout, _ = proc.communicate()
        print(proc.args)
        print(stdout.decode(encoding='utf-8'))

    def wait(self):
        while True:
            for proc in self.__processes:
                ret = proc.poll()
                if ret is not None:
                    self.__processes.remove(proc)
                    if ret != 0:
                        self.__running = False
                        Executor.__process_failed(proc)
                        self.wait_all()
                        exit(1)

    def wait_all(self):
        for proc in self.__processes:
            proc.wait()
            if proc.returncode != 0:
                self.__running = False
                Executor.__process_failed(proc)
        self.__processes = []



from setuptools import setup, find_packages
from os import path


root_dir = path.abspath(path.dirname(__file__))
package_name = 'irori'


def _requirements():
    return [name.rstrip() for name in open(path.join(root_dir, 'requirements.txt')).readlines()]


setup(
    name='irori',
    version='1.0.0',
    license='GPLv3.0',
    packages=find_packages(),

    install_requires=_requirements(),

    author='SiLeader',
    url='https://github.com/SiLeader/irori',

    description='general purpose built tool',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: Japanese',
        'Natural Language :: English',

        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: Microsoft :: Windows :: Windows 7',
        'Operating System :: Microsoft :: Windows :: Windows 8',
        'Operating System :: Microsoft :: Windows :: Windows 8.1',
        'Operating System :: POSIX',

        'Topic :: Software Development :: Build Tools'
    ]
)



import enum
import platform
import os


class Os(enum.Enum):
    Linux = 'linux'
    Darwin = 'darwin'
    Windows = 'windows'
    Others = 'others'


__os = None


def _resolve_os():
    global __os
    if __os is None:
        pf = platform.system()
        pf = pf.lower()
        if 'windows' in pf:
            __os = Os.Windows
        elif 'linux' in pf:
            __os = Os.Linux
        elif 'darwin' in pf:
            __os = Os.Darwin
        else:
            __os = Os.Others
    return __os


class Context:
    @property
    def os(self):
        return _resolve_os()

    @property
    def is_posix(self):
        return os.name.lower() == 'posix'

    @property
    def is_windows(self):
        return self.os == Os.Windows

    @property
    def is_linux(self):
        return self.os == Os.Linux

    @property
    def is_darwin(self):
        return self.os == Os.Darwin
